"""nodoctest
all.py -- much of sage is imported into this module, so you don't
          have to import everything individually. 
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


import os, sys
if sys.version_info[:2] < (2, 4):
    print >>sys.stderr, "Sorry, SAGE requires Python 2.4 or later"
    sys.exit(1)


###################################################################
from sage.libs.all       import *
from sage.rings.all      import *
from sage.matrix.all     import *
from sage.modules.all    import *
from sage.monoids.all    import *
from sage.algebras.all   import *
from sage.misc.all       import *
from sage.modular.all    import *
from sage.schemes.all    import *
from sage.modular.all    import *
from sage.groups.all     import *
from sage.databases.all  import *
from sage.structure.all  import *
from sage.categories.all import *
from sage.sets.all       import *
from sage.interfaces.all import *
from sage.functions.transcendental import *
import sage.tests.all as tests

###################################################################

# maximize memory resources
try:
    import resource   # unix only...
    resource.setrlimit(resource.RLIMIT_AS, (-1,-1))
except:
    pass

# abreviations
EC = EllipticCurve
MS = ModularSymbols
VS = VectorSpace
CS = cuspidal_subspace
NS = new_subspace
oo = infinity
x  = PolynomialRing(RationalField(),'x').gen()


def OS(x):
    return x.old_subspace()

def ES(x):
    return x.eisenstein_subspace()

def NCMS(n,k=2,sign=1):
    return ms(n,weight=k,sign=sign).cuspidal_subspace().new_subspace()

tm = misc.cputime

QQ = RationalField()
CC = ComplexField()
I = CC.gen(0)
i = CC.gen(0)

# These are used by the pre-parser, and should not
# be overwritten by the user, unless they want to change
# the meaning of int and real in the interpreter (which
# is a valid thing to do, and doesn't mess up anything
# else in SAGE).
ZZ = IntegerRing()
RR = RealField()  # default real field

# Some shorter shortcuts, which the user can safely
# overwrite.
Q = QQ
Z = ZZ
R = RR
C = CC

# grab signal handling back from PARI or other C libraries
get_sigs()

# Error message that matches the SAGE/IPython defaults
quit = "Use Ctrl-D (i.e. EOF), %Exit, or %Quit to exit without confirmation."
exit = quit

iglob = set(['Out', 'In', 'help', 'view_all', \
             'os', 'iglob', 'variables']).union(set(globals().keys()))
