"""
Coercion helper functions
"""

#*****************************************************************************
#       Copyright (C) 2004-2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import __builtin__
import operator

cimport _element
import  _element

cimport _module
import  _module

def lcm(x,y):
    return x.lcm(y)

def gcd(x,y):
    return x.gcd(y)

def xgcd(x,y):
    return x.xgcd(y)

def parent(x):
    try:
        return x.parent()
    except AttributeError:
        return type(x)

def coerce(p, x):
    try:
        return p._coerce_(x)
    except AttributeError:
        return p(x)
            

def canonical_coercion(x, y):
    cdef int i
    xp = parent(x)
    yp = parent(y)
    if xp is yp:
        return x, y
    
    if isinstance(x, (int, long)):
        x = yp(x)
    elif isinstance(y, (int, long)):
        y = xp(y)
    else:
        i = 0
        try:
            x = coerce(yp, x)
        except TypeError, ValueError:
            i = i + 1
        try:
            y = coerce(xp, y)
        except TypeError, ValueError:
            i = i + 1
        if i == 0:
            raise TypeError, "unable to find an unambiguous parent for %s (parent: %s) and %s (parent: %s)"%(x,xp, y, yp)
        elif i == 2:
            import  sage.ext._ring
            if isinstance(x, sage.ext._ring.Ring) or isinstance(y, sage.ext._ring.Ring):
                raise TypeError, "you cannot add ring (=%s) to a number or to another ring (=%s)!"%(x, y)
            raise TypeError, "unable to find any common parent for %s (parent: %s) and %s (parent: %s)"%(x,xp, y, yp)            
    return x, y


def bin_op(x, y, op):
    if isinstance(y, _element.InfinityElement):
        return op(y,x)
    if op == operator.mul and \
           isinstance(y, (\
                      _element.ModuleElement,
                      _element.AlgebraElement,
                      _module.Module)) and \
           isinstance(x, (_element.RingElement, int, long, float)):
        return op(y,x)
    x, y = canonical_coercion(x, y)
    return op(x,y)

        
def cmp(x, y):
    if type(x) == type(None) and type(y) != type(None):
        return -1
    elif type(x) != type(None) and type(y) == type(None):
        return -1
    elif isinstance(y, _element.InfinityElement):
        return -y.__cmp__(x)
    try:
        x, y = canonical_coercion(x, y)
    except TypeError:
        return -1  # not equal, since can't coerce.
    return __builtin__.cmp(x,y)
    #return x.__cmp__(y)
