"""nodoctest
Permutation Groups

WARNING: *NOT* done -- many of the examples below don't work
         *exactly* as indicated etc.  Work in progress.

AUTHOR: David Joyner, first version 2005-10-14

disjoint cycle notation:

(a,...,b)(c,...,d)...(e,...,f)  <--> [[a,...,b],[c,...,d],...,[e,...,f]]
                  () = identity <--> [] or [[]]

EXAMPLES:
    sage: G = PermutationGroup([[[1,2,3],[4,5]],[[3,4]]])
    sage: G
    Permutation Group of degree 5 with generators [(1, 2, 3)(4, 5), (3, 4)]
    sage: G.group_id() 
    [120, 34]
"""

import sage.structure.element as element
import sage.structure.gens
import sage.rings.rational_field as rationals
import sage.misc.misc 
import sage.misc.functional as fun
import sage.rings.arith as arith
import sage.matrix.matrix_space as matrix
import random

from sage.interfaces.all import gap as Gap

# Uncommenting the following would give us our own interface
# to GAP; it's more efficient to just use the SAGE-wide one.
#import sage.interfaces.gap 
#_gap = None
#def Gap(cmd):
#    global _gap
#    if _gap is None:
#        _gap = sage.interfaces.gap.GAP()
#    return _gap(cmd)

def list_string_parser(s):
    """
    Input: s is the string repn of a list of things
    
    Output: returns a python list of strings.

    EXAMPLES:
        sage: s = "[1,2,37]"
        sage: list_string_parser(s)
        ['1', '2', '37']
    """
    L=list(s)
    n=len(L)
    i0=L.index(",")
    ans0=[s[1:i0]]
    for i in range(i0+1,len(L)):
        if L[i]==",":
            ans0.append(s[i0:i])
            i0=i
    ans0.append(s[i0:n-1])
    ans=[]
    for i in range(len(ans0)):
        ans.append(ans0[i].replace(",",""))
    ans0,ans=ans,[]
    for i in range(len(ans0)):
        ans.append(ans0[i].replace(" ",""))
    return ans
	
	
def perm_value_cyclic(self, x):
    """
    self is a permutation in disjoint cycle notation
    x>0 is an integer.computes self(x)

    EXAMPLES:
        sage: A = [[1,2,3,4,5]]
        sage: pg.perm_value_cyclic(A,5)
        1
        sage: pg.perm_value_cyclic(A,22)
        22
    """ 
    perm=self[0]
    n=len(perm)
    try:
        j=perm.index(x)
        if j<n-1:
            return perm[j+1]
        else:
            return perm[0]
    except ValueError:
        return x
    
def perm_value(self, x):
    """
    self is a permutation in disjoint cycle notation
    x>0 is an integer.computes self(x)

    EXAMPLES:
        sage: A=[[1,2,3],[4,5,6]]
        sage: pg.perm_value(A,3)
        1
        sage: pg.perm_value(A,5)
        6
        sage: pg.perm_value(A,12)
        12
    """
    n=len(self)
    x0=x
    if n==1:
        return perm_value_cyclic(self, x)
    for j in range(n):
        x0=perm_value_cyclic([self[j]], x0)
    return x0
    

def perm_orbit(self,x):
    """
	 self is a permutation of [1,2,...,n]
         x is in 1,2,...,n
         returns the orbit of x under the powers of the
         associated permutation in disjoint cycle notation

    EXAMPLES:
        sage: perm=[3,1,2]
        sage: orb=pg.perm_orbit(perm,1); orb
        [1, 3, 2]
    """
    perm=self
    n=max(perm)          
    orb=[x]
    val=x
    for k in range(n):
        i=int(val-1)
	#print k,val,perm,i,perm[i],"\n"
	val=perm[i]      
        #orb.append(val)
        #print k,val,"\n"
        if val==x:
            break
            print x,val,val!=x,"\n"
        if val!=x:
            orb.append(val)
    return orb

def perm2disjcyc(self):
    """
	 self is a permutation of [1,2,...,n]
         returns the associated permutation in
         disjoint cycle notation


    EXAMPLES:
        sage: import sage.groups.permgroup as pg
        sage: A=[1,2,5,4,3]
        sage: pg.perm2disjcyc(A)
        [[3, 5]]
    """
    perm=self
    n=len(perm)
    cyc=[]
    used=[]
    for j in range(1,n+1):
        val=perm[j-1]
        if j!=val and not(j in used):
            orb=perm_orbit(perm,j)
            #print "orb: ",orb
            cyc.append(orb)
            used=used+orb
            #print j,orb
        if val==j:
            used=used+[j]
    #print used
    return cyc

def disjcyc2perm(self,deg):
    """
	inverse to perm2disjcyc

    EXAMPLES:
        sage: A=[1,2,5,4,3]
        sage: a=pg.perm2disjcyc(A)
        sage: a
        [[3, 5]]
        sage: pg.disjcyc2perm(a,5)
        [1, 2, 5, 4, 3]
    """
    L=[perm_value(self,x) for x in range(1,deg+1)]
    return L

###########################################################################
##      built in function sample in random module
##	Description:    -Unique Random Numbers
##                      -Returns a list of specified number of elements from
##                       another list such that each element in the returned
##                       list is unique and random.
##                       See e.g. below.
##
## from random import *
##>>> sample([1,2,3,4],4)
##[1, 3, 2, 4]
##>>> sample([1,2,3,4],4)
##[4, 1, 2, 3]
###########################################################################


def random_dcn(deg):
    """
    returns a random element of S_deg in
    disjoint cycle notation (DCN)

    EXAMPLES:
        sage.: pg.random_dcn(5)
        [[1, 2], [3, 4, 5]]
    """
    r=random.sample(range(1,deg+1),deg)
    return perm2disjcyc(r)

def order(self):
    """
    self is a permutation in DCN

    EXAMPLES:
        sage: G=pg.PermGroup([[[1,2,3]]],3)
        sage: gens=G.generators()
        sage: g=pg.PermGroupElement(G,gens[0])
        sage: g
        Permutation Group Element (1, 2, 3)
        sage: g.order()
        3
    """
    lens=[len(self[i]) for i in range(len(self))]
    return arith.LCM(lens)

def mult_dcn(self,other):
    """
    self, other are permutations in disjoint cycle notation

    EXAMPLES:
        sage: A=[[1,2,3,4,5]]; B=[[1,2,3],[4,5,6]]
        sage: pg.mult_dcn(A,B)
        [[1, 3, 5, 2], [4, 6]]
    """
    lperm=self
    rperm=other
    if lperm==[] or lperm==[[]]:
        return rperm
    if rperm==[] or rperm==[[]]:
	return lperm
    n=max(sum(lperm,[])+sum(rperm,[]))
    perm=[perm_value(rperm,perm_value(lperm,i)) for i in range(1,n+1)]
    #print perm
    return perm2disjcyc(perm)

def power_dcn(self,exp):
    """
    self is a permutation in disjoint cycle notation
    exp is an integer

    EXAMPLES:
        sage: A=[[1,2,3,4,5]]; B=[[1,2,3],[4,5]]
        sage: pg.order(A)
        5
        sage: pg.order(B)
        6
        sage: pg.power_dcn(B,4)
        [[1, 2, 3]]
        sage: pg.power_dcn(B,5)
        [[1, 3, 2], [4, 5]]
        sage: pg.power_dcn(A,4)
        [[1, 5, 4, 3, 2]]
        sage: pg.power_dcn(A,-1)
        [[1, 5, 4, 3, 2]]
        sage: pg.power_dcn(A,0)
        []	
    """
    if exp==1:
	return self
    if exp==0:
	return []
    count=0
    ans=[]
    if exp>0:
	for i in range(exp):
       	    ans=mult_dcn(ans,self)
        return ans
    if exp<0:
	inv=power_dcn(self,order(self)-1)
	for i in range(-exp):
	    ans=mult_dcn(ans,inv)
        return ans

def swap(L):
    """
    permutation L in list notation
    returns Bruhat length or swapping nmuber
    the "sign of perm" is = (-1)^swap(perm)

    EXAMPLES:
        sage: A=[[1,2,3,4,5]]; B=[[1,2,3],[4,5]]
        sage: a=pg.disjcyc2perm(A,5); b=pg.disjcyc2perm(B,5); a; b
        [2, 3, 4, 5, 1]
        [2, 3, 1, 5, 4]
        sage: pg.swap(a); pg.swap(b)
        4
        3
    """
    n=len(L)
    temp=0
    for i1 in range(n-1):
        j1=L[i1]
        for i2 in range(i1,n):
            j2=L[i2]
            if j1>j2:
                temp=temp+1
    return temp

def sign_perm(self):
    """
    self is a permutation in DCN
    computes the sign of self = (-1)^swap(self)

    EXAMPLES:
        sage: A=[[1,2,3,4,5]]; B=[[1,2,3],[4,5]]
        sage: a = pg.disjcyc2perm(A,5); b = pg.disjcyc2perm(B,5); a; b
        [2, 3, 4, 5, 1]
        [2, 3, 1, 5, 4]
        sage: pg.sign_perm(A); pg.sign_perm(B)  
        1
        -1
    """
    deg=max(sum(self,[]))
    n=swap(disjcyc2perm(self,deg))
    if fun.is_odd(n):
       	return -1
    return 1
    #return (-1)^n   ###### doesn't work!!!

def perm2mat(self,deg):
    """
    Returns deg x deg permutation matrix associated
    to the permutation self

    EXAMPLES:
        sage: pg.perm2mat(A,5)
        [0 1 0 0 0]
        [0 0 1 0 0]
        [0 0 0 1 0]
        [0 0 0 0 1]
        [1 0 0 0 0]
    """
    M = matrix.MatrixSpace(rationals.RationalField(),deg,deg)
    A = M(0)
    for i in range(deg):
	A[i,perm_value(self,i+1)-1]=1
    return A

class PermutationGroup(sage.structure.gens.Generators):
    """

    EXAMPLES:
        sage: G = pg.PermutationGroup([[[1,2,3],[4,5]],[[3,4]]])
        sage: G
        Permutation Group of degree 5 with generators [(1, 2, 3)(4, 5), (3, 4)]
        sage: G.center()
        Permutation Group of degree 5 with generators []
        sage: G.group_id()
        [120, 34]
        sage: n = G.order()
        sage: n
        120

        sage.: r = G.random()
        sage.: r
        Permutation Group Element (2, 5)
    """
    def __init__(self, gens, deg=0):
        self.__gens = gens
	if deg==0:
	    self.__degree = max(sum(sum(self.__gens,[]),[]))
	else:
	    self.__degree=deg

    def gens(self):
        return self.__gens

    def degree(self):
        return self.__degree
    
    def __repr__(self):
        deg=self.__degree
        gens=str(self.__gens)
	gens=gens.replace("[[[","[(")
	gens=gens.replace("]]]",")]")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("[[","(")
	gens=gens.replace("]]",")")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("]], [[","),(")
	gens=gens.replace("],[",")(")
	gens=gens.replace("], [",")(")
        #gens=gens.replace("]",")")
	#gens=gens.replace("[","(")
        return "Permutation Group of degree "+str(deg)+" with generators %s"%gens

    def order(self):
        gens=str(self.__gens)
	# start parsing (SAGE->GAP)
	gens=gens.replace("[[[","[(")
	gens=gens.replace("]]]",")]")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("[[","(")
	gens=gens.replace("]]",")")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("]], [[","),(")
	gens=gens.replace("],[",")(")
	gens=gens.replace("], [",")(")
	i="G:=Group( " + gens + ");"
        o = Gap(i)
        #print "\n",i," -> ",o,"\n "
	i="ans:=Size(G);"
	o = Gap(i)
        #print "\n",i," -> ",o,"\n"
        i="ans;"
	o = Gap(i)
        #print "\n",i," -> ",o,"\n "
	result=o
	ans=int(result)
        return ans

    def random(self):
        """
        This calls GAP's Random(G) function
	A reverse parse needed
        """
        gens=str(self.__gens)
	# start parsing (SAGE->GAP)
	gens=gens.replace("[[[","[(")
	gens=gens.replace("]]]",")]")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("[[","(")
	gens=gens.replace("]]",")")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("]], [[","),(")
	gens=gens.replace("],[",")(")
	gens=gens.replace("], [",")(")
	i="G:=Group( " + gens + ")"
	o = Gap(i)
	i="r:=Random(G)"
	o = Gap(i)
	i="ans:=ListPerm(r)"
	o = Gap(i)
	res0=o
	# start reverse parsing (GAP->SAGE)
        res=res0.replace(")(","],[")
        res=res0.replace(")","]")
        res=res.replace("(","[")
	res1=list_string_parser(res0)
	res=[]
	for i in range(len(res1)):
	    res.append(int(res1[i]))    ## this is a SAGE list of integers
		                        ## representing a permutation
        return PermutationGroupElement(self,perm2disjcyc(res))

    def __contains__(self,x):
        # call GAP's "in" function
	gens=str(self.gens())
	# start parsing (SAGE->GAP)
	gens=gens.replace("[[[","[(")
	gens=gens.replace("]]]",")]")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("[[","(")
	gens=gens.replace("]]",")")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("]], [[","),(")
	gens=gens.replace("],[",")(")
	gens=gens.replace("], [",")(")
	i="G:=Group( " + gens + ")"
	o = Gap(i)
	g=str(x.dcn())
	g=g.replace("[[[","[(")
	g=g.replace("]]]",")]")
	g=g.replace("]],[[","),(")
	g=g.replace("[[","(")
	g=g.replace("]]",")")
	g=g.replace("]],[[","),(")
	g=g.replace("]], [[","),(")
	g=g.replace("],[",")(")
	g=g.replace("], [",")(")
	i=g+" in G"
	o = Gap(i)
	res=o
	print "\n "+res+" \n"
	if res=="true":
	    return 1
	return 0

    def group_id(self):
	# call GAP's "IdGroup" function
	gens=str(self.gens())
	# start parsing (SAGE->GAP)
	gens=gens.replace("[[[","[(")
	gens=gens.replace("]]]",")]")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("[[","(")
	gens=gens.replace("]]",")")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("]], [[","),(")
	gens=gens.replace("],[",")(")
	gens=gens.replace("], [",")(")
	i="G:=Group( " + gens + ")"
	o = Gap(i)
	i="IdGroup(G)"
	o=Gap(i)
	return eval(o)

    def center(self):
        # call GAP's "Center" function
	deg=max(sum(sum(self.gens(),[]),[]))	    
	gens=str(self.gens())
	# start parsing (SAGE->GAP)
	gens=gens.replace("[[[","[(")
	gens=gens.replace("]]]",")]")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("[[","(")
	gens=gens.replace("]]",")")
	gens=gens.replace("]],[[","),(")
	gens=gens.replace("]], [[","),(")
	gens=gens.replace("],[",")(")
	gens=gens.replace("], [",")(")
	i="G:=Group( " + gens + ")"
	o = Gap(i)
	i="H:=Center(G)"
	o=Gap(i)
	i="gens:=GeneratorsOfGroup(H)"
	o=Gap(i)
	gens=o
	gens=gens.replace("),(","]], [[")
	gens=gens.replace("), (","]], [[")
	gens=gens.replace(")(","], [")
	gens=gens.replace("("," [")
	gens=gens.replace(")","] ")
	gens=eval(gens)
	return PermutationGroup(gens,deg)

class PermutationGroupElement(element.MultiplicativeGroupElement):
    """
    An element of a permutation group.

    Create elements by first defining the finite group
    G using gens gens, then use the notation
    G.gens() and multiplication * to find the product.

    EXAMPLES:
        sage: G = pg.PermutationGroup([[[1,2,3],[4,5]]],3)
        sage: G
        Permutation Group with generators  [(1, 2, 3)(4, 5)]
        sage: gens=G.gens()
        sage: g=pg.PermutationGroupElement(G,gens[0])
        sage: g
        Permutation Group Element (1, 2, 3)(4, 5)
        sage: print g
        (1, 2, 3)(4, 5)
        sage: g*g
        Permutation Group Element (1, 3, 2)
        sage: G = pg.PermutationGroup([[[1,2,3]]],3)
        sage: gens = G.gens()
        sage: g = pg.PermutationGroupElement(G,gens[0])
        sage: g
        Permutation Group Element (1, 2, 3)
        sage: g.order()
        3
    """
    def __init__(self, parent,g):
        """
        Create element of a permutation group.
        """
        element.Element.__init__(self, parent)  
        self.__parent = parent
        self.__element = g

    def __repr__(self):
        return "Permutation Group Element %s"%self.__str__()

    def __str__(self):
        s=str(self.__element)
	#s=s.replace("[[","(")
	#s=s.replace("]]",")")
	#s=s.replace("]],[[","),(")
	#s=s.replace("]], [[","),(")
	s=s.replace("],[",")(")
	s=s.replace("], [",")(")
        s=s.replace("]]",")")
	s=s.replace("[[","(")
	return s
	

    def __getitem__(self, i):
        # self[i]
        return int(perm_value(self.__element,int(i))) 

    def dcn(self):
        #return the disjoint cycle notation in list form
        return self.__element

    def __mul__(self,other):
        # self, other are elements of a PermutationGroup 
        lperm=self.__element
        rperm=other.__element
        n=max(sum(lperm,[])+sum(rperm,[]))
        perm=[perm_value(rperm,perm_value(lperm,i)) for i in range(1,n+1)]
        #print "\n",perm,"\n",perm2disjcyc(perm),"\n"
	return PermutationGroupElement(self.__parent, perm2disjcyc(perm))

    def order(self):
        perm=self.__element
        lens=[len(perm[i]) for i in range(len(perm))]
        return arith.LCM(lens)

    def __len__(self):
        return len(self.order())        # len(self)


##################################################################
## Code from email  "Re: [Numpy-discussion] Permutation in Numpy"
## by David M. Cooke (Jul 28 2004 11:05PM)
## Computes the list of all permutations of a list, k
## One is able to access them in order, so you're only
## storing the current permutation instead of all of them,
## saving memory.
##>>> L=perm([1,2,3,4])
##>>> L.next()
##(1, 2, 3, 4)
################################################################
def perm(k):
    k = tuple(k)
    lk = len(k)
    if lk <= 1:
        yield k
    elif lk == 2:
        yield k
        yield (k[1], k[0])
    elif lk == 3:
        k0, k1, k2 = k
        yield k
        yield (k0, k2, k1)
        yield (k1, k0, k2)
        yield (k1, k2, k0)
        yield (k2, k0, k1)
        yield (k2, k1, k0)
    else:
        for i in range(lk):
            s = k[:i] + k[i+1:]
            t = (k[i],)
            for x in perm(s):
                yield t + x


#######################################################

permgroup_testcommands="""
import sage.groups.permgroup as pg
pg.load_cas("gap")  ####### speeds up some commands if this is pre-loaded
G=pg.PermutationGroup([[[1,2,3],[4,5]],[[3,4]]])
G
G.center()
G.group_id() # this hangs in 0.7.11 -- ERROR
r=G.random()
r
n=G.order()
n
G=pg.PermutationGroup([[[1,2,3],[4,5]]],3)
G
gens=G.gens()
g=pg.PermutationGroupElement(G,gens[0])
g
print g
g*g
G=pg.PermutationGroup([[[1,2,3]]],3)
gens=G.gens()
g=pg.PermutationGroupElement(G,gens[0])
g
g.order()
perm=[3,1,2]
orb=pg.perm_orbit(perm,1); orb
A=[[1,2,3,4,5]]
pg.perm_value_cyclic(A,5)
pg.perm_value_cyclic(A,22)
A=[[1,2,3],[4,5,6]]
pg.perm_value(A,3)
pg.perm_value(A,5)
pg.perm_value(A,12)
A=[[1,2,3,4,5]]; B=[[1,2,3],[4,5,6]]
pg.mult_dcn(A,B)
A=[[1,2,3,4,5]]; B=[[1,2,3],[4,5]]
pg.order(A)
pg.order(B)
pg.power_dcn(B,4)
pg.power_dcn(B,5)
pg.power_dcn(A,4)
pg.power_dcn(A,-1)
pg.power_dcn(A,0)
a=pg.disjcyc2perm(A,5); b=pg.disjcyc2perm(B,5); a; b
pg.swap(a); pg.swap(b)
pg.sign_perm(A); pg.sign_perm(B) 
pg.perm2mat(A,5)
pg.random_dcn(5)
"""


if __name__ ==  '__main__':
    import doctest, sys
    import permgroup as pg
    doctest.testmod(sys.modules[__name__])
