r"""
Interface to MAGMA

\note{You must have \code{magma} installed on your computer
for this interface to work.   MAGMA is not free, so it is
not included with SAGE, but you can buy it at
         http://magma.maths.usyd.edu.au}
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from expect import Expect
PROMPT = ">>>"

class Magma(Expect):
    def __init__(self, maxread=10000, script_subdirectory="user"):
        Expect.__init__(self,
                        name = "magma",
                        prompt = ">",
                        command = "magma -b",
                        maxread = maxread,
                        script_subdirectory = script_subdirectory,
                        restart_on_ctrlc = True)

    def __call__(self, x):
        x = str(x).rstrip()
        if len(x) == 0 or x[len(x) - 1] != ';':
            x += ';'
        return Expect.__call__(self, x)

    def _start(self):
        self._change_prompt('>')
        Expect._start(self)
        self('SetPrompt("%s");SetLineEditor(false);'%PROMPT)
        self._change_prompt(PROMPT)
        self.expect().expect(PROMPT)
        self.expect().expect(PROMPT)

    def set(self, var, value):
        """
        Set the variable var to the given value.
        """
        self("%s := %s"%(var, value))

    def get(self, var):
        """
        Get the value of the variable var.
        """
        return self("%s"%var)

    def clear(self, var):
        """
        Clear the variable named var.
        """
        self("delete %s"%var)


magma = Magma()

import os
def magma_console():
    os.system('magma')

