"""
Interface to the Octave interpreter
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import os

from expect import Expect

from sage.misc.misc import verbose

class Octave(Expect):
    """
    Send the code x to the Octave interpreter and return the output as
    a string.
    """
    def __init__(self, maxread=100, script_subdirectory=""):
        Expect.__init__(self,
                        name = 'octave',
                        prompt = '>',
                        command = "octave --no-line-editing --silent",
                        maxread = maxread,
                        script_subdirectory = script_subdirectory,
                        restart_on_ctrlc = False,
                        verbose_start = False)

    def _start(self):
        Expect._start(self)
        self("page_screen_output=0;")
        self("format none;")
        #self('save_header_format_string="";')

    def get_via_file(self, var_name):
        t = self._temp_file(var_name)
        self('save -text "%s" %s'%(t,var_name))
        r = open(t).read()
        os.unlink(t)
        return r

    def set_via_file(self, var_name, x):
        t = self._temp_file(var_name)
        open(t,'w').write(x)
        print 'load "%s" %s'%(t, var_name)        
        self('load "%s" %s'%(t, var_name))
        #os.unlink(t)

# An instance
octave = Octave(script_subdirectory='user')

import os
def octave_console():
    os.system('octave')
