"""
Format SAGE documentation for viewing with IPython
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import re

substitutes = [('$',''), ('\\',''), ('sage.:', 'sage:'), ('backslash','\\'), \
               ('begin{enumerate}',''), ('end{enumerate}',''), ('mapsto', ' |--> '), \
               ('ldots', '...')]

def _rmcmd(s, cmd, left='', right=''):
    c = '\\%s{.*}'%cmd
    r = re.compile(c, re.DOTALL)
    while True:
        m = r.search(s)
        if m is None: break
        s = s[:m.start()] + left + s[m.start()+len(cmd)+1:m.end()-1] \
            + right + s[m.end():]
    return s
        
def format(s):
    """
    Format SAGE documentation for viewing with IPython.
    """
    assert isinstance(s, str)
    s = _rmcmd(s, 'code')
    s = _rmcmd(s, 'emph')
    for a,b in substitutes:
        s = s.replace(a,b)
    return s
