"""
Abelain monoid elements

AUTHOR: David Kohel, 2005-09
"""

#*****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful, 
#    but WITHOUT ANY WARRANTY; without even the implied warranty 
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# 
#  See the GNU General Public License for more details; the full text 
#  is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

__doc_exclude = ['Integer', 'MonoidElement']

import operator

from sage.ext._integer import Integer
from sage.ext._element import MonoidElement

def is_FreeAbelianMonoidElement(x):
    return isinstance(x, FreeAbelianMonoidElement)

class FreeAbelianMonoidElement(MonoidElement):
    def __init__(self, F, x):
        """
        Create the element x of the FreeAbelianMonoid F.
        """
        MonoidElement.__init__(self, F)
        self.__repr = None
        n = F.ngens()
        if isinstance(x, (int, Integer)) and x == 1:
            self.__element_vector = [ 0 for i in range(n) ]
        elif isinstance(x, list):
            if len(x) != n: 
                raise IndexError, \
                      "Argument length (= %s) must be %s."%(len(x), n)
            self.__element_vector = x
        else:
            raise TypeError, "Argument x (= %s) is of wrong type."%x
                                        
    def __repr__(self):
        s = ""
        A = self.parent()
        n = A.ngens()
        x = A.names()
        v = self.__element_vector
        for i in range(n):
            if v[i] == 1: 
                if len(s) > 0: s += "*"
                s += "%s"%x[i]
            elif v[i] != 0: 
                if len(s) > 0: s += "*"
                s += "%s^%s"%(x[i],v[i])
        if len(s) == 0: s = "1"
        return s

    def __mul__(self, y):
        if not isinstance(y, FreeAbelianMonoidElement):
            raise TypeError, "Argument y (= %s) is of wrong type."%y
        M = self.parent()
        z = M(1)
        xelt = self.__element_vector
        yelt = y.__element_vector
        z.__element_vector = [ xelt[i]+yelt[i] for i in range(len(xelt)) ]
        return z

    def __pow__(self, n):
        if not isinstance(n, (int, long, Integer)):
            raise TypeError, "Argument n (= %s) must be an integer."%n
        if n < 0: 
            raise IndexError, "Argument n (= %s) must be positive."%n
        elif n == 0:
            return self.parent()(1)
        elif n == 1:
            return self
        elif n == 2:
            return self * self
        k = n//2
        return self**k * self**(n-k)

    def list(self):
        """
        Return (a reference to) the underlying list used to represent
        this element.  If this is a monoid in an abelian monoid on $n$
        generators, then this is a list of nonnegative integers of
        length $n$.
        """
        return self.__element_vector



# doctests
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
