"""
Complex numbers
"""

import complex_number
import field

_obj_CC = {}
class _uniq_CC(object):
    def __new__(cls):
        if _obj_CC.has_key(0):
            return _obj_CC[0]
        O = object.__new__(cls)
        _obj_CC[0] = O
        return O

def is_ComplexField(x):
    return isinstance(x, ComplexField)
    
class ComplexField(_uniq_CC, field.Field):
    """
    The field of complex numbers.
    """

    def __init__(self):
        """
            EXAMPLES:
            sage: C = ComplexField(); C
            Complex Field
            sage: C('1/3')
            0.333333333333333333333
            sage: C('1/3', 2)
            0.333333333333333333333 + 2.0*I

        Note that the second argument is the number of *bits* of precision,
        not the number of digits of precision:
            sage: C('1/3',2, 100)
            0.3333333333333333333333333333333333333333 + 2.0*I

        We can also coerce rational numbers and integers into C, but
        coercing a polynomial in raising an exception.
            sage: Q = RationalField()
            sage: C(Q('1/3'))
            0.333333333333333333333
            sage: C(Q('1/3'), Q('2/3'), 10)
            0.333333333333333333333 + 0.666666666666666666667*I
            sage: S = PolynomialRing(Q)
            sage: C(S.gen())
            Traceback (most recent call last):
            ...
            TypeError: unable to coerce (x,0) to a ComplexNumber
            sage: C.is_field()
            True
            sage: C.characteristic()
            0
            sage: C.name()
            'C'
            sage: C == C
            True
            sage: C == 1.1
            False
        """
        pass
    
    def __call__(self, re, im=None, bits=53):
        if isinstance(re, complex_number.ComplexNumber) and (bits == 0 or re.prec() == bits):
            return re
        return complex_number.ComplexNumber(re, im, bits)
    
    def __init__(self):
        pass
        
    def __repr__(self):
        return "Complex Field"

    def characteristic(self):
        return 0

    def gen(self, n=0, bits=53):
        if n != 0:
            raise AttributeError, "Generator must be 0"
        return complex_number.ComplexNumber(0,1,bits)
     
    def is_field(self):
        return True
     
    def ngens(self):
        return 1


        
CC = ComplexField()

# doctests
if __name__ ==  '__main__':
    import doctest, sys
    from rational_field import RationalField
    from polynomial_ring import PolynomialRing
    doctest.testmod(sys.modules[__name__])
