from infinity_element import InfinityElement
from ring_element import RingElement




class _uniq(object):
    """
    Anything that derives from this class will have at most
    one instance.
    """
    _obj = None
    def __new__(cls):
        if cls._obj != None:
            return cls._obj
        cls._obj = InfinityElement.__new__(cls, None)
        return cls._obj

class Infinity(_uniq, InfinityElement):
    def __init__(self):
        InfinityElement.__init__(self, None)
    
    def __repr__(self):
        return "Infinity"

    def __add__(self, other):
        if isinstance(other, (RingElement, Infinity, int, long)):
            return self
        raise TypeError, "%s + %s not defined."%(self, other)

    def __radd__(self, left):
        return self+left
    
    def __sub__(self, other):
        if isinstance(other, (RingElement, int, long)) and not isinstance(other, InfinityElement):
            return self
        raise TypeError, "%s - %s not defined."%(self, other)        
    
    def __rsub__(self, other):
        if isinstance(other, (RingElement, int, long)) and not isinstance(other, InfinityElement):
            return self
        raise TypeError, "%s - %s not defined."%(other, self)
    
    def __mul__(self, other):
        if isinstance(other, Infinity):
            return self
        if isinstance(other, (RingElement, int, long)):
            if other == 0:
                return other
            return self
        raise TypeError, "%s * %s not defined."%(self*other)
    
    def __rmul__(self, left):
        return self*left
    
    def __div__(self, other):
        if isinstance(other, (RingElement, int, long)) and not isinstance(other, InfinityElement):
            if other != 0:
                return self
        raise TypeError, "%s * %s not defined."%(self*other)
    
    def __cmp__(self, other):
        if isinstance(other, InfinityElement):
            return 0
        return 1  # infinity is bigger than everything but itself


infinity = Infinity()

