"""
Laurent series ring
"""

import weakref

from rational_field import Q
import laurent_series
import power_series_ring
import commutative_ring
import integral_domain
import field


_objsLaurentSeriesRing = {}
def LaurentSeriesRing(base_ring, name=None):
    global _objsLaurentSeriesRing
    key = (base_ring, name)
    if _objsLaurentSeriesRing.has_key(key):
        x = _objsLaurentSeriesRing[key]()
        if x != None: return x

    if isinstance(base_ring, field.Field):
        R = LaurentSeriesRing_field(base_ring, name)
    elif isinstance(base_ring, integral_domain.IntegralDomain):
        R = LaurentSeriesRing_domain(base_ring, name)
    elif isinstance(base_ring, commutative_ring.CommutativeRing):
        R = LaurentSeriesRing_generic(base_ring, name)
    else:
        raise TypeError, "base_ring must be a commutative ring"
    _objsLaurentSeriesRing[key] = weakref.ref(R)
    return R

def is_LaurentSeriesRing(x):
    return isinstance(x, LaurentSeriesRing_generic)

class LaurentSeriesRing_generic(commutative_ring.CommutativeRing):
    """
    Univariate Laurent Series Ring
    """

    def __init__(self, base_ring, name=None):
        self.__base_ring = base_ring
        self.assign_names(name)

    def __repr__(self):
        return "Laurent series ring in %s over %s"%(self.name(), self.base_ring())
    
    def __call__(self, x, n=0):
        if isinstance(x, laurent_series.LaurentSeries) and n==0 and self == x.parent():
            return x
        return laurent_series.LaurentSeries(self, x, n)
    
    def __cmp__(self, other):
        if not isinstance(other, LaurentSeriesRing_generic):
            return -1
        if self.base_ring() != other.base_ring():
            return -1
        if self.name() != other.name():
            return -1
        return 0

    def base_ring(self):
        return self.__base_ring

    def gen(self, n=0):
        if n != 0:
            raise IndexError, "Generator %s not defined."%n
        try:
            return self.__generator
        except AttributeError:
            self.__generator = laurent_series.LaurentSeries(self, [0,1])
            return self.__generator

    def ngens(self):
        return 1

    def power_series_ring(self):
        """
        If this is the Laurent series ring R((t)), return the power
        series ring R[[t]].
        """
        try:
            return self.__power_series_ring
        except AttributeError:
            self.__power_series_ring = power_series_ring.PowerSeriesRing(
                                         self.base_ring(), self.name())
            return self.__power_series_ring            

class LaurentSeriesRing_domain(integral_domain.IntegralDomain, LaurentSeriesRing_generic):
    def __init__(self, base_ring, name=None):
        LaurentSeriesRing_generic.__init__(self, base_ring, name)

    def fraction_field(self):
        try:
            return self.__fraction_field
        except AttributeError:
            self.__fraction_field = LaurentSeriesRing(self.base_ring().fraction_field(), name)
            return self.__fraction_field
    
class LaurentSeriesRing_field(field.Field, LaurentSeriesRing_generic):
    def __init__(self, base_ring, name=None):
        LaurentSeriesRing_generic.__init__(self, base_ring, name)

