"""
Quotient of arbitrary ring by an ideal.

AUTHOR: William Stein
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import quotient_ring_element
import sage.misc.latex as latex
import commutative_ring
import ideal

def QuotientRing(R, I):
    """
    Create the quotient of R by the ideal I.

    INPUT:
        R -- a commutative ring
        I -- an ideal of R

    EXAMPLES:
    """
    if not isinstance(R, commutative_ring.CommutativeRing):
        raise TypeError, "R must be a commutative ring."
    if not isinstance(I, ideal.Ideal_generic) or I.ring() != R:
        I = R.ideal(I)
    return QuotientRing_generic(R, I)
    
class QuotientRing_generic(commutative_ring.CommutativeRing):
    """
    The quotient ring of R by the ideal I.
    """
    def __init__(self, R, I):
        """
        Create the quotient ring of R by the ideal I.
        INPUT:
            R -- a commutative ring
            I -- an ideal
        """
        self.__R = R
        self.__I = I

    def __repr__(self):
        return "Quotient of %s by the ideal %s"%(self.ring(), self.ideal())

    def _latex_(self):
        return "%s/%s"%(latex.latex(self.ring()), latex.latex(self.ideal()))

    def characteristic(self):
        raise NotImplementedError
    
    def ideal(self):
        return self.__I
        
    def is_field(self):
        return self.ideal().is_maximal()

    def is_integral_domain(self):
        return self.ideal().is_prime()

    def ring(self):
        return self.__R
    
    def __call__(self, x, coerce=True):
        if isinstance(x, quotient_ring_element.QuotientRingElement):
            if x.parent() == self:
                return x
            x = x.lift()
        if coerce:
            R = self.ring()
            x = R(x)
        return quotient_ring_element.QuotientRingElement(self, x, coerce=False)
    
    def __cmp__(self, other):
        if not isinstance(other, QuotientRing_generic):
            return -1
        if self.ring() == other.ring() and self.ideal() is other.ideal():
            return 0
        return 1

    def ngens(self):
        return self.ring().ngens()

    def gen(self, i=0):
        return quotient_ring_element.QuotientRingElement(self, self.ring().gen(i))

    
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
