"""
Generic Set object
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.structure.element import Element
import sage.categories.all

def Set(X):
    if is_Set(X):
        return X

    if isinstance(X, Element):
        raise TypeError, "Element has no defined underlying set"
    #try:
    #    if X.is_finite():
    #        return Set_finite(X)
    #except AttributeError:
    #    pass
    return Set_generic(X)

def is_Set(x):
    return isinstance(x, Set_generic)
    

class Set_generic:
    def __init__(self, X):
        self.__object = X

    def __repr__(self):
        return "Set of elements of %s"%self.__object

    def __iter__(self):
        return self.__object.__iter__()

    def __contains__(self, x):
        return x in self.__object

    def __cmp__(self, X):
        if is_Set(X) and self.__object == X.__object:
            return 0
        return -1

    def union(self, X):
        if is_Set(X):
            if self == X:
                return self
            return Set_union(self.__object, X.__object)
        raise TypeError, "X must be a Set"

    def intersection(self, X):
        if is_Set(X):
            if self == X:
                return self
            return Set_intersection(self.__object, X.__object)
        raise TypeError, "X must be a Set"


    def __len__(self):
        try:
            return len(self.__object)
        except TypeError:
            raise NotImplementedError, "computation of order of %s not yet implemented"%self.__object

    def order(self):
        return len(self)

    def category(self):
        return sage.categories.all.Sets()

    def object(self):
        return self.__object


class EnumeratedSet(Set_generic):
    def __init__(self, X):
        Set_generic.__init__(self, X)

    def __iter__(self):
        for x in self.set():
            yield x

    def __repr__(self):
        s = str(self.set())
        return "{" + s[5:-2] + "}"
        #    return "Finite set of elements of %s"%self.__object

    def __contains__(self, x):
        return x in self.set()


    def set(self):
        try:
            return self.__set
        except AttributeError:
            self.__set = set(self.object())
            return self.__set

    def __cmp__(self, other):
        if isinstance(other, Set_finite):
            if self.set() == other.set():
                return 0
            else:
                return -1
        else:
            return Set_generic.__cmp__(self, other)

    def union(self, other):
        if not isinstance(other, Set_finite):
            return Set_generic.union(self, other)
        return Set_finite(self.set().union(other.set()))

    def intersection(self, other):
        if not isinstance(other, Set_finite):
            return Set_generic.intersection(self, other)
        return Set_finite(self.set().intersection(other.set()))
    

class Set_union(Set_generic):
    def __init__(self, X, Y):
        self.__X = X
        self.__Y = Y
        Set_generic.__init__(self, self)

    def __repr__(self):
        return "Set-theoretic union of %s and %s"%(self.__X, self.__Y)

    def __iter__(self):
        for x in self.__X:
            yield x
        for y in self.__Y:
            yield y

    def __contains__(self, x):
        return x in self.__X or x in self.__Y

    def order(self):
        return self.__X.order() + self.__Y.order()

class Set_intersection(Set_generic):
    def __init__(self, X, Y):
        self.__X = X
        self.__Y = Y
        Set_generic.__init__(self, self)

    def __repr__(self):
        return "Set-theoretic intersection of %s and %s"%(self.__X, self.__Y)

    def __iter__(self):
        for x in self.__X:
            if x in self.__Y:
                yield x

    def __contains__(self, x):
        return x in self.__X and x in self.__Y

    def order(self):
        raise NotImplementedErro

