"""
Abstract base class for elements.
"""

##################################################################
# Generic element, so all this functionality must be defined
# by any element.  Derived class must call __init__
##################################################################

import sage.structure.element_py

# This classes uses element.pxd.  To add data members, you
# must change that file. 

cdef class Element:
    """
    Generic element.
    """

    def __init__(self, parent, required_instance=None):
        self.__parent = parent
        if not (required_instance is None) and not isinstance(parent, required_instance):
            raise TypeError, "Parent must be of instance %s but it is of type %s"%(required_instance, type(parent))

    def __reduce__(self):
        raise NotImplementedError, "object persistence not yet supported for this element"

    def __repr__(self):
        return "Generic element of a ring"

    def __hash__(self):
        return hash(str(self))

    def __xor__(self, right):
        raise RuntimeError, "Use ** for exponentiation, not '^', which means xor\n"+\
              "in Python, and has the wrong precedence."

    def base_ring(self):
        return self.parent().base_ring()

    def is_zero(self):
        return self == self.parent()(0)
    
    def parent(self, x=None):
        if x==None:
            return self.__parent
        else:
            return self.__parent(x)

    def order(self):
        raise NotImplementedError

    def divides(self, other):
        """
        Return True if self divides other.
        """
        try:
            return self.parent().divides(self, other)
        except AttributeError:
            raise NotImplementedError


