"""
Type that holds the factorization of an object
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <wstein@ucsd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import sage.misc.latex as latex

class Factorization(list):
    def __init__(self, x):
        if not isinstance(x, list):
            raise TypeError, "x must be a list"
        from sage.rings.integer import Integer
        for t in x:
            if not (isinstance(t, tuple) and len(t) == 2 and isinstance(t[1],(int, long, Integer))):
                raise TypeError, "x must be a list of tuples of length 2"
        list.__init__(self, x)
        self.sort()

    def __repr__(self):
        if len(self) == 0:
            return '(1)'
        try:
            atomic = ((isinstance(self[0][0], (int, long)) or \
                       self[0][0].parent().is_atomic_repr()))
        except AttributeError:
            atomic = False
        s = ''
        for i in range(len(self)):
            t = str(self[i][0])
            if not atomic  and (t.find('+') != -1 or t.find('-') != -1):
                t = '(%s)'%t
            n = self[i][1]
            if n != 1:
                t += '^%s'%n
            s += t
            if i < len(self)-1:
                s += ' * '
        return s

    def _latex_(self):
        if len(self) == 0:
            return '(1)'
        try:
            atomic = ((isinstance(self[0][0], (int, long)) or \
                       self[0][0].parent().is_atomic_repr()))
        except AttributeError:
            atomic = False
        s = ''
        for i in range(len(self)):
            t = latex.latex(self[i][0])
            if not atomic and (t.find('+') != -1 or t.find('-') != -1):
                t = '(%s)'%t
            n = self[i][1]
            if n != 1:
                t += '^{%s}'%latex.latex(n)
            s += t
            if i < len(self)-1:
                s += ' \\cdot '
        return s

    def __add__(self, other):
        raise NotImplementedError

    def __mul__(self, other):
        if not isinstance(other, Factorization):
            raise TypeError
        d1 = dict(self)
        d2 = dict(other)
        s = {}
        for a in set(d1.keys()).union(set(d2.keys())):
            s[a] = 0
            if d1.has_key(a):
                s[a] += d1[a]
            if d2.has_key(a):
                s[a] += d2[a]
        return Factorization(list(s.iteritems()))
