#!/usr/bin/env python
DEVEL = False

import distutils.sysconfig, os, sys
from distutils.core import setup, Extension


NO_WARN = True

if not os.environ.has_key('SAGE_HOME'):
    print "    ERROR: The environment variable SAGE_HOME must be defined."
    sys.exit(0)

if not os.environ.has_key('SAGE_VERSION'):
    print "    ERROR: The environment variable SAGE_VERSION must be defined.\n\
    It doesn't really matter what it is defined to."
    sys.exit(0)


SAGE_LOCAL = os.environ['SAGE_HOME'] + '/local/'
SAGE_VERSION = os.environ['SAGE_VERSION']

ec =    Extension('sage.libs._ec.ec',
              sources = ["sage/libs/_ec/ec.pyx"] +  \
                        ["sage/libs/_ec/%s"%s for s in \
                         ["analrank.c", "apcompute.c", "arderivs.c",
                          "arintern.c", "arith.c", "artwists.c",
                          "arutil.c", "checkit.c", "degphi.c", 
                          "diskio.c", "docurve.c", "dodisk.c", 
                          "equation.c", "exotic.c", "fixit.c", 
                          "iisog2.c", "iisog3.c", "isog.c", "isog2.c",
                          "isog23.c", "isog24.c", "isog3.c", "isog5.c",
                          "isog52.c", "isog713.c", "isogNprime.c",
                          "isoggen.c", "isogsort.c", "isogx0.c", 
                          "isogx0branch.c", "isogx0branch1.c", 
                          "isogx0branch2.c", "isogx0branch3.c",
                          "isogx0branch4.c", "isogx0branch5.c", 
                          "isogx0branch6.c", "isogx0getd.c", 
                          "isogx0period.c", "readit.c",
                          "special.c", "util.c"]], 
              libraries = ["pari", "m"])

hanke = Extension(name = "sage.libs._hanke.hanke",
              sources = ["sage/libs/_hanke/hanke.pyx",
                         "sage/libs/_hanke/wrap.cc",
                         "sage/libs/_hanke/Matrix_mpz/Matrix_mpz.cc",
                         "sage/libs/_hanke/Matrix_mpz/CountLocal2.cc",
                         "sage/libs/_hanke/Matrix_mpz/CountLocal.cc",
                         "sage/libs/_hanke/Matrix_mpz/Local_Constants.cc",
                         "sage/libs/_hanke/Matrix_mpz/Local_Density_Front.cc",
                         "sage/libs/_hanke/Matrix_mpz/Local_Density_Congruence.cc",
                         "sage/libs/_hanke/Matrix_mpz/Local_Normal.cc",
                         "sage/libs/_hanke/Matrix_mpz/Local_Invariants.cc",
                         "sage/libs/_hanke/Utilities/string_utils.cc",
                         "sage/libs/_hanke/GMP_class_extras/mpz_class_extras.cc",
                         "sage/libs/_hanke/GMP_class_extras/vectors.cc" ],
                   libraries = ["gmp", "gmpxx", "stdc++"])

ntl_modules = [Extension('sage.libs._ntl.ntl_ZZ',
                         sources = ["sage/libs/_ntl/ntl_ZZ.pyx"]), \
               Extension('sage.libs._ntl.ntl_ZZ_p',
                         sources = ["sage/libs/_ntl/ntl_ZZ_p.pyx"]), \
               Extension('sage.libs._ntl.ntl_ZZX',
                         sources = ["sage/libs/_ntl/ntl_ZZX.pyx"]), \
               Extension('sage.libs._ntl.ntl_ZZ_pX',
                         sources = ["sage/libs/_ntl/ntl_ZZ_pX.pyx"]), \
               Extension('sage.libs._ntl.ntl_mat_ZZ',
                         sources = ["sage/libs/_ntl/ntl_mat_ZZ.pyx"])]

for m in ntl_modules:
    m.libraries = ["ntl", "gmp", "gmpxx", "m", "stdc++"]
    m.sources.append("sage/libs/_ntl/wrap.cc")

mwrank =  Extension("sage.libs._mwrank.mwrank",
              sources = ["sage/libs/_mwrank/mwrank.pyx",
                         "sage/libs/_mwrank/wrap.cc"],
              define_macros = [("NTL_ALL",None)],
              libraries = ["mwrank", "ntl", "gmp", "gmpxx", "stdc++", "m", "pari"])

pari_modules = [Extension('sage.libs._pari.gen',
                          sources = ["sage/libs/_pari/gen.pyx"])]

if DEVEL:
    pari_modules.append(Extension('sage.libs._pari.test',
                          sources = ["sage/libs/_pari/test.pyx"]))
    
for m in pari_modules:
    m.libraries += ['pari', 'gmp']


ext_modules = [ \
    ec, \
    mwrank, \

    Extension('sage.ext._arith',
              sources = ['sage/ext/_arith.pyx']), \

    Extension('sage.ext._arith_gmp',
              sources = ['sage/ext/_arith_gmp.pyx'],
              libraries=['gmp']), \

    Extension('sage.ext.coerce',
              sources = ['sage/ext/coerce.pyx']), \

    Extension('sage.ext._congroup_pyx',
              sources = ['sage/ext/_congroup_pyx.pyx', \
                         'sage/ext/_arith.pyx']), \

    Extension('sage.ext._element',
              sources = ['sage/ext/_element.pyx']), \

    Extension('sage.ext._module',
              sources = ['sage/ext/_module.pyx']), \

    Extension('sage.ext._ring',
              sources = ['sage/ext/_ring.pyx']), \

    Extension('sage.ext._gens',
              sources = ['sage/ext/_gens.pyx']), \

    Extension('sage.ext._mpfr',
              sources = ['sage/ext/_mpfr.pyx', 'sage/ext/_ring.pyx'],
              libraries = ['mpfr', 'gmp']), \

    Extension('sage.ext._mpfr',
              sources = ['sage/ext/_mpfr.pyx', 'sage/ext/_ring.pyx'],
              libraries = ['mpfr', 'gmp']), \

    Extension('sage.ext._integer',
              sources = ['sage/ext/_arith.pyx', 'sage/ext/_integer.pyx'],
              libraries=['gmp']), \

    Extension('sage.ext._intmod_pyx',
              sources = ['sage/ext/_intmod_pyx.pyx']), \

    Extension('sage.ext._polynomial_pyx',
              sources = ['sage/ext/_polynomial_pyx.pyx',
                         'sage/ext/_arith_gmp.pyx'],
              libraries=['gmp']), \

    Extension('sage.ext._rational',
              sources = ['sage/ext/_rational.pyx',
                         'sage/ext/_arith.pyx', \
                         'sage/ext/_integer.pyx'],
              libraries=['gmp']), \

    Extension('sage.ext._sparse_poly', 
              sources = ['sage/ext/_sparse_poly.c'],
              libraries=['gmp']), \

    Extension('sage.ext._sparse_matrix_pyx',
              ['sage/ext/_sparse_matrix_pyx.pyx',
               'sage/ext/_integer.pyx',
               'sage/ext/_rational.pyx',
               'sage/ext/_arith.pyx'],
              libraries=['gmp']), \

    Extension('sage.ext._dense_matrix_pyx',
              ['sage/ext/_dense_matrix_pyx.pyx',
               'sage/ext/_integer.pyx',
               'sage/ext/_rational.pyx',
               'sage/ext/_arith.pyx'],
              libraries=['gmp']), \

    Extension('sage.ext._search',
              ['sage/ext/_search.pyx']), \

    Extension('sage.ext._heilbronn',
              ['sage/ext/_heilbronn.pyx',
               'sage/ext/_p1list.pyx',
               'sage/ext/_arith.pyx'],
              libraries = ['gmp', 'm']), \

    Extension('sage.ext._p1list',
              ['sage/ext/_p1list.pyx',
               'sage/ext/_arith.pyx'],
              libraries = ['gmp']), \
    ] + \
    ntl_modules + \
    pari_modules

mpc =Extension('sage.ext._mpc',
              sources = ['sage/ext/_mpc.pyx', 'sage/ext/_ring.pyx'],
              libraries = ['mpc', 'mpfr', 'gmp'])


if NO_WARN and distutils.sysconfig.get_config_var('CC').startswith("gcc"):
    extra_compile_args = ['-w']

if DEVEL:
    extra_compile_args.append('-ggdb')
    ext_modules.append(hanke)
    ext_modules.append(mpc)

for m in ext_modules:
    m.sources += ['sage/libs/interrupt.c']

include_dirs = ['%s/include'%SAGE_LOCAL,
                '%s/include/python2.4'%SAGE_LOCAL]

extra_link_args =  ['-L%s/lib'%SAGE_LOCAL]

def pyrex(ext_modules):
    for m in ext_modules:
        m.extra_compile_args += extra_compile_args
        m.extra_link_args += extra_link_args
        for i in range(len(m.sources)):
            f = m.sources[i]
            if f[-4:] == ".pyx":
                m.sources[i] = f[:-4] + ".c"
                c_file_exists = os.path.exists("%s.c"%f[:-4])
                if not c_file_exists or \
                       (c_file_exists and 
                        os.path.getctime("%s.c"%f[:-4]) <= os.path.getctime(f)):
                    cmd = "pyrexc %s"%f
                    print cmd
                    n = os.system(cmd)
                    if n != 0:
                        print "Error running pyrex2c."
                        sys.exit(0)

pyrex(ext_modules)


setup(name='sage',

      version     =  SAGE_VERSION,
      
      description = 'SAGE: System for Algebra and Geometry Experimentation',

      license     = 'GNU Public License (GPL)',
      
      author      = 'William Stein',

      author_email= 'was@math.harvard.edu',
      
      url         = 'http://sage.sourceforge.net',
      
      packages    = ['sage',
                     'sage.algebras',
                     'sage.categories',
                     'sage.databases',
                     'sage.ext',
                     'sage.functions',
                     'sage.groups',
                     'sage.interfaces',
                     'sage.libs',
                     'sage.libs._hanke',
                     'sage.libs._mwrank',
                     'sage.libs._ntl',
                     'sage.libs._ec',
                     'sage.libs._pari',
                     'sage.matrix',
                     'sage.misc',
                     'sage.modules',
                     'sage.modular',
                     'sage.modular.abvar',
                     'sage.modular.hecke',
                     'sage.modular.modform',
                     'sage.modular.modsym',
                     'sage.modular.ssmod',
                     'sage.monoids',
                     'sage.plot',
                     'sage.rings',
                     'sage.sets',
                     'sage.tests',
                     'sage.schemes',
                     'sage.schemes.affine_schemes',
                     'sage.schemes.affine_space',
                     'sage.schemes.group_schemes',
                     'sage.schemes.group_schemes.abelian_varieties',
                     'sage.schemes.group_schemes.abelian_varieties.jacobians',
                     'sage.schemes.group_schemes.abelian_varieties.modular',
                     'sage.schemes.group_schemes.finite_flat',
                     'sage.schemes.group_schemes.linear_groups',
                     'sage.schemes.group_schemes.neron_models',
                     'sage.schemes.hypersurfaces',
                     'sage.schemes.hypersurfaces.kummer_surfaces',
                     'sage.schemes.hypersurfaces.plane_curves',
                     'sage.schemes.hypersurfaces.plane_curves.elliptic',
                     'sage.schemes.projective_schemes',
                     'sage.schemes.projective_space',
                     'sage.structure',
                     ],
      
      ext_modules = ext_modules,
      include_dirs = include_dirs)

 
 
